@@ -40,6 +40,10 @@ class OsalNvIds(BaseNvIds):
     HAS_CONFIGURED_ZSTACK1 = 0x0F00
     HAS_CONFIGURED_ZSTACK3 = 0x0060
 
+    # Although the docs say "IDs reserved for applications range from 0x0401 to 0x0FFF",
+    # no OSAL NVID beyond 0x03FF is writable with the MT interface when using Z-Stack 3.
+    ZIGPY_ZNP_MIGRATION_ID = 0x005F
+
     # OSAL NV item IDs
     EXTADDR = 0x0001
     BOOTCOUNTER = 0x0002
@@ -199,7 +203,7 @@ class OsalNvIds(BaseNvIds):
     TCLK_DEFAULT = 0x0103  # Not accually a Nv Item but Id used by SecMgr
 
     LEGACY_TCLK_IC_TABLE_START = 0x0104  # Deprecated. Refer to EX_TCLK_IC_TABLE
-    LEGACY_TCLK_IC_TABLE_END = 0x0110  # IC keys, refered with shift byte
+    LEGACY_TCLK_IC_TABLE_END = 0x0110  # IC keys, referred with shift byte
 
     LEGACY_TCLK_TABLE_START = 0x0111  # Deprecated. Refer to EX_TCLK_TABLE
     LEGACY_TCLK_TABLE_END = 0x01FF
