@@ -2,6 +2,7 @@ from __future__ import annotations
 
 import os
 import time
+import random
 import asyncio
 import logging
 import itertools
@@ -16,6 +17,7 @@ import zigpy.device
 import async_timeout
 import zigpy.endpoint
 import zigpy.profiles
+import zigpy.zdo.types as zdo_t
 import zigpy.application
 import zigpy.zcl.foundation
 from zigpy.zcl import clusters
@@ -23,10 +25,12 @@ from zigpy.types import ExtendedPanId, deserialize as list_deserialize
 from zigpy.zdo.types import CLUSTERS as ZDO_CLUSTERS, ZDOCmd, ZDOHeader, MultiAddress
 from zigpy.exceptions import DeliveryError
 
+import zigpy_znp.const as const
 import zigpy_znp.types as t
 import zigpy_znp.config as conf
 import zigpy_znp.commands as c
 from zigpy_znp.api import ZNP
+from zigpy_znp.znp import security
 from zigpy_znp.utils import combine_concurrent_calls
 from zigpy_znp.exceptions import CommandNotRecognized, InvalidCommandResponse
 from zigpy_znp.types.nvids import OsalNvIds
@@ -40,7 +44,6 @@ STARTUP_TIMEOUT = 5
 ZDO_REQUEST_TIMEOUT = 15
 DATA_CONFIRM_TIMEOUT = 8
 DEVICE_JOIN_MAX_DELAY = 5
-NETWORK_COMMISSIONING_TIMEOUT = 30
 WATCHDOG_PERIOD = 30
 BROADCAST_SEND_WAIT_DURATION = 3
 MULTICAST_SEND_WAIT_DURATION = 3
@@ -60,25 +63,15 @@ REQUEST_TRANSIENT_ERRORS = {
 
 REQUEST_ROUTING_ERRORS = {
     t.Status.APS_NO_ACK,
+    t.Status.APS_NOT_AUTHENTICATED,
     t.Status.NWK_NO_ROUTE,
+    t.Status.NWK_INVALID_REQUEST,
     t.Status.MAC_NO_ACK,
     t.Status.MAC_TRANSACTION_EXPIRED,
 }
 
 REQUEST_RETRYABLE_ERRORS = REQUEST_TRANSIENT_ERRORS | REQUEST_ROUTING_ERRORS
 
-Z2M_PAN_ID = 0xA162
-Z2M_EXT_PAN_ID = t.EUI64.convert("DD:DD:DD:DD:DD:DD:DD:DD")
-Z2M_NETWORK_KEY = t.KeyData([1, 3, 5, 7, 9, 11, 13, 15, 0, 2, 4, 6, 8, 10, 12, 13])
-
-DEFAULT_TC_LINK_KEY = t.TCLinkKey(
-    ExtAddr=t.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"),  # global
-    Key=t.KeyData(b"ZigBeeAlliance09"),
-    TxFrameCounter=0,
-    RxFrameCounter=0,
-)
-ZSTACK_CONFIGURE_SUCCESS = t.uint8_t(0x55)
-
 LOGGER = logging.getLogger(__name__)
 
 
@@ -100,9 +93,7 @@ class ZNPCoordinator(zigpy.device.Device):
             model = "CC2538" if self.application._znp.nvram.align_structs else "CC2531"
             version = "Home 1.2" if self.application._znp.version == 1.2 else "3.0.x"
 
-        build = self.application._version_rsp.CodeRevision
-
-        return f"{model}, Z-Stack {version} (build {build})"
+        return f"{model}, Z-Stack {version} (build {self.application._zstack_build_id})"
 
 
 class ControllerApplication(zigpy.application.ControllerApplication):
@@ -131,20 +122,6 @@ class ControllerApplication(zigpy.application.ControllerApplication):
     # Implementation of the core zigpy ControllerApplication methods #
     ##################################################################
 
-    @property
-    def network_key(self) -> t.KeyData | None:
-        # This is not a standard Zigpy property
-        if self.state.network_information.network_key:
-            return self.state.network_information.network_key.key
-        return None
-
-    @property
-    def network_key_seq(self) -> t.uint8_t | None:
-        # This is not a standard Zigpy property
-        if self.state.network_information.network_key:
-            return self.state.network_information.network_key.seq
-        return None
-
     @classmethod
     async def probe(cls, device_config: conf.ConfigType) -> bool:
         """
@@ -212,6 +189,9 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         self._znp = znp
         self._znp.set_application(self)
 
+        if not read_only and not force_form:
+            await self._migrate_nvram()
+
         self._bind_callbacks()
 
         # Next, read out the NVRAM item that Zigbee2MQTT writes when it has configured
@@ -225,9 +205,10 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             configured_value = await self._znp.nvram.osal_read(
                 configured_nv_item, item_type=t.uint8_t
             )
-            is_configured = configured_value == ZSTACK_CONFIGURE_SUCCESS
         except KeyError:
             is_configured = False
+        else:
+            is_configured = configured_value == const.ZSTACK_CONFIGURE_SUCCESS
 
         if force_form:
             LOGGER.info("Forming a new network")
@@ -264,7 +245,6 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator)
         )
 
-        # The AUTOSTART startup NV item doesn't do anything.
         if self._znp.version == 1.2:
             # Z-Stack Home 1.2 has a simple startup sequence
             await self._znp.request(
@@ -310,7 +290,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         if self.znp_config[conf.CONF_LED_MODE] is not None:
             await self._set_led_mode(led=0xFF, mode=self.znp_config[conf.CONF_LED_MODE])
 
-        await self._load_network_info()
+        await self.load_network_info()
         await self._register_endpoints()
 
         # Setup the coordinator as a zigpy device and initialize it to request node info
@@ -328,7 +308,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         LOGGER.info("Network settings")
         LOGGER.info("  Model: %s", self.zigpy_device.model)
         LOGGER.info("  Z-Stack version: %s", self._znp.version)
-        LOGGER.info("  Z-Stack build id: %s", self._version_rsp.CodeRevision)
+        LOGGER.info("  Z-Stack build id: %s", self._zstack_build_id)
         LOGGER.info("  Max concurrent requests: %s", max_concurrent_requests)
         LOGGER.info("  Channel: %s", self.channel)
         LOGGER.info("  PAN ID: 0x%04X", self.pan_id)
@@ -336,57 +316,19 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         LOGGER.info("  Device IEEE: %s", self.ieee)
         LOGGER.info("  Device NWK: 0x%04X", self.nwk)
         LOGGER.debug(
-            "  Network key: %s", ":".join(f"{c:02x}" for c in self.network_key)
+            "  Network key: %s",
+            ":".join(
+                f"{c:02x}" for c in self.state.network_information.network_key.key
+            ),
         )
 
-        if self.network_key == Z2M_NETWORK_KEY:
+        if self.state.network_information.network_key.key == const.Z2M_NETWORK_KEY:
             LOGGER.warning(
                 "Your network is using the insecure Zigbee2MQTT network key!"
             )
 
         self._watchdog_task = asyncio.create_task(self._watchdog_loop())
 
-    async def update_network_channel(self, channel: t.uint8_t):
-        """
-        Changes the network channel, increments the beacon update ID, and waits until
-        the changes take effect.
-
-        Does nothing if the new channel is the same as the old.
-        """
-
-        if self.channel == channel:
-            return
-
-        await self._znp.request(
-            request=c.ZDO.MgmtNWKUpdateReq.Req(
-                Dst=0x0000,
-                DstAddrMode=t.AddrMode.NWK,
-                Channels=t.Channels.from_channel_list([channel]),
-                ScanDuration=0xFE,  # switch channels
-                ScanCount=0,
-                NwkManagerAddr=0x0000,
-            ),
-            RspStatus=t.Status.SUCCESS,
-        )
-
-        # The above command takes a few seconds to work
-        while self.channel != channel:
-            await self._load_network_info()
-            await asyncio.sleep(1)
-
-    async def update_pan_id(self, pan_id: t.uint16_t) -> None:
-        """
-        Updates the network PAN ID, bypassing Z-Stack anti-collision checks.
-        """
-
-        await self._znp.reset()
-
-        nib = await self._znp.nvram.osal_read(OsalNvIds.NIB, item_type=t.NIB)
-        nib.nwkPanId = pan_id
-
-        await self._znp.nvram.osal_write(OsalNvIds.NIB, nib)
-        await self._znp.nvram.osal_write(OsalNvIds.PANID, pan_id)
-
     async def set_tx_power(self, dbm: int) -> None:
         """
         Sets the radio TX power.
@@ -420,13 +362,9 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         network_key = self.config[conf.CONF_NWK][conf.CONF_NWK_KEY]
 
         if pan_id is None:
-            # Let Z-Stack pick one at random, hopefully not conflicting with others
-            pan_id = t.uint16_t(0xFFFF)
+            pan_id = random.SystemRandom().randint(0x0001, 0xFFFE + 1)
 
         if extended_pan_id is None:
-            # It's not documented whether or not Z-Stack will pick this randomly as well
-            # if a value of 00:00:00:00:00:00:00:00 is provided but the chances of a
-            # collision using `os.urandom` are astronomically small
             extended_pan_id = ExtendedPanId(os.urandom(8))
 
         if network_key is None:
@@ -436,150 +374,36 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         if channel is not None:
             channels = t.Channels.from_channel_list([channel])
 
-        # Delete any existing HAS_CONFIGURED_ZSTACK* NV items. One (or both) may fail.
-        await self._znp.nvram.osal_delete(OsalNvIds.HAS_CONFIGURED_ZSTACK1)
-        await self._znp.nvram.osal_delete(OsalNvIds.HAS_CONFIGURED_ZSTACK3)
-        await self._znp.nvram.osal_delete(OsalNvIds.BDBNODEISONANETWORK)
-
-        # Instruct Z-Stack to reset everything on the next boot
-        await self._znp.nvram.osal_write(
-            OsalNvIds.STARTUP_OPTION,
-            t.StartupOptions.ClearState | t.StartupOptions.ClearConfig,
+        network_info = zigpy.state.NetworkInformation(
+            extended_pan_id=extended_pan_id,
+            pan_id=pan_id,
+            nwk_update_id=self.config[conf.CONF_NWK][conf.CONF_NWK_UPDATE_ID],
+            nwk_manager_id=0x0000,
+            channel=channel,
+            channel_mask=channels,
+            security_level=5,
+            network_key=zigpy.state.Key(
+                key=network_key,
+                tx_counter=0,
+                rx_counter=0,
+                seq=0,
+                partner_ieee=None,
+            ),
+            tc_link_key=None,
+            children=[],
+            key_table=[],
+            nwk_addresses={},
+            stack_specific={"zstack": {"tclk_seed": os.urandom(16).hex()}},
         )
 
-        # And reset to clear everything
-        await self._znp.reset()
-
-        # Now that we've cleared everything, write back our Z-Stack settings
-        LOGGER.debug("Updating network settings")
+        node_info = zigpy.state.NodeInfo(
+            nwk=0x0000,
+            ieee=None,
+            logical_type=zdo_t.LogicalType.Coordinator,
+        )
 
+        await self.write_network_info(network_info=network_info, node_info=node_info)
         await self._write_stack_settings(reset_if_changed=False)
-
-        for item, value in {
-            # Ignore the user's PAN ID for now and just let Z-Stack pick one that works
-            OsalNvIds.PANID: t.uint16_t(0xFFFF),
-            OsalNvIds.APS_USE_EXT_PANID: extended_pan_id,
-            # XXX: Z2M incorrectly uses this NVRAM item, APS_USE_EXT_PANID is correct
-            OsalNvIds.EXTENDED_PAN_ID: extended_pan_id,
-            OsalNvIds.PRECFGKEY: network_key,
-            # XXX: Z2M requires this item to be False
-            OsalNvIds.PRECFGKEYS_ENABLE: t.Bool(False),
-            OsalNvIds.CHANLIST: channels,
-        }.items():
-            await self._znp.nvram.osal_write(item, value, create=True)
-
-        # Z-Stack Home 1.2 doesn't have the BDB subsystem
-        if self._znp.version > 1.2:
-            await self._znp.request(
-                c.AppConfig.BDBSetChannel.Req(IsPrimary=True, Channel=channels),
-                RspStatus=t.Status.SUCCESS,
-            )
-            await self._znp.request(
-                c.AppConfig.BDBSetChannel.Req(
-                    IsPrimary=False, Channel=t.Channels.NO_CHANNELS
-                ),
-                RspStatus=t.Status.SUCCESS,
-            )
-
-        # Finally, form the network
-        LOGGER.debug("Forming the network")
-
-        started_as_coordinator = self._znp.wait_for_response(
-            c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator)
-        )
-
-        # Handle the startup progress messages
-        async with self._znp.capture_responses(
-            [
-                c.ZDO.StateChangeInd.Callback(
-                    State=t.DeviceState.StartingAsCoordinator
-                ),
-                c.AppConfig.BDBCommissioningNotification.Callback(
-                    partial=True,
-                    Status=c.app_config.BDBCommissioningStatus.InProgress,
-                ),
-            ]
-        ):
-            try:
-                if self._znp.version > 1.2:
-                    # Z-Stack 3 uses the BDB subsystem
-                    commissioning_rsp = await self._znp.request_callback_rsp(
-                        request=c.AppConfig.BDBStartCommissioning.Req(
-                            Mode=c.app_config.BDBCommissioningMode.NwkFormation
-                        ),
-                        RspStatus=t.Status.SUCCESS,
-                        callback=c.AppConfig.BDBCommissioningNotification.Callback(
-                            partial=True,
-                            RemainingModes=c.app_config.BDBCommissioningMode.NONE,
-                        ),
-                        timeout=NETWORK_COMMISSIONING_TIMEOUT,
-                    )
-
-                    if (
-                        commissioning_rsp.Status
-                        != c.app_config.BDBCommissioningStatus.Success
-                    ):
-                        raise RuntimeError(
-                            f"Network formation failed: {commissioning_rsp}"
-                        )
-                else:
-                    await self._znp.nvram.osal_write(
-                        OsalNvIds.TCLK_SEED,
-                        value=DEFAULT_TC_LINK_KEY,
-                        create=True,
-                    )
-
-                    # In Z-Stack 1.2.2, StartupFromApp actually does what it says
-                    await self._znp.request(
-                        c.ZDO.StartupFromApp.Req(StartDelay=100),
-                        RspState=c.zdo.StartupState.NewNetworkState,
-                    )
-
-                # Both versions still end with this callback
-                await started_as_coordinator
-            except asyncio.TimeoutError as e:
-                raise RuntimeError(
-                    "Network formation refused, RF environment is likely too noisy."
-                    " Temporarily unscrew the antenna or shield the coordinator"
-                    " with metal until a network is formed."
-                ) from e
-
-        LOGGER.debug("Waiting for the NIB to be populated")
-
-        # Even though the device says it is "ready" at this point, it takes a few more
-        # seconds for `_NIB.nwkState` to switch from `NwkState.NWK_INIT`. There does
-        # not appear to be any user-facing MT command to read this information.
-        while True:
-            try:
-                nib = await self._znp.nvram.osal_read(OsalNvIds.NIB, item_type=t.NIB)
-
-                LOGGER.debug("Current NIB is %s", nib)
-
-                # Usually this works after the first attempt
-                if nib.nwkLogicalChannel != 0 and nib.nwkPanId != 0xFFFE:
-                    break
-            except KeyError:
-                pass
-
-            await asyncio.sleep(1)
-
-        if pan_id != 0xFFFF:
-            # Update the PAN ID at runtime after the network has formed to make sure
-            # Z-Stack does not complain about collisions on Z-Stack 3.
-            await self.update_pan_id(pan_id)
-
-        # Create the NV item that keeps track of whether or not we're fully configured.
-        if self._znp.version == 1.2:
-            configured_nv_item = OsalNvIds.HAS_CONFIGURED_ZSTACK1
-        else:
-            configured_nv_item = OsalNvIds.HAS_CONFIGURED_ZSTACK3
-
-        await self._znp.nvram.osal_write(
-            configured_nv_item, ZSTACK_CONFIGURE_SUCCESS, create=True
-        )
-
-        # Finally, reset once more to reset the device back to a "normal" state so that
-        # we can continue the normal application startup sequence.
         await self._znp.reset()
 
     def get_dst_address(self, cluster):
@@ -704,10 +528,11 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         # If joins were permitted through a specific router, older Z-Stack builds
         # did not allow the key to be distributed unless the coordinator itself was
-        # also permitting joins.
+        # also permitting joins. This also needs to happen if we're permitting joins
+        # through the coordinator itself.
         #
         # Fixed in https://github.com/Koenkk/Z-Stack-firmware/commit/efac5ee46b9b437
-        if time_s == 0 or self._version_rsp.CodeRevision < 20210708:
+        if time_s == 0 or self._zstack_build_id < 20210708 or node == self.ieee:
             response = await self._znp.request_callback_rsp(
                 request=c.ZDO.MgmtPermitJoinReq.Req(
                     AddrMode=t.AddrMode.NWK,
@@ -979,6 +804,14 @@ class ControllerApplication(zigpy.application.ControllerApplication):
     # Internal methods #
     ####################
 
+    @property
+    def _zstack_build_id(self) -> t.uint32_t:
+        """
+        Z-Stack build ID, more recently the build date.
+        """
+
+        return self._version_rsp.CodeRevision
+
     @property
     def zigpy_device(self) -> zigpy.device.Device:
         """
@@ -1101,6 +934,58 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         except (asyncio.TimeoutError, CommandNotRecognized):
             LOGGER.info("This build of Z-Stack does not appear to support LED control")
 
+    async def _migrate_nvram(self):
+        """
+        Migrates NVRAM entries using the `ZIGPY_ZNP_MIGRATION_ID` NVRAM item.
+        """
+
+        try:
+            migration_id = await self._znp.nvram.osal_read(
+                OsalNvIds.ZIGPY_ZNP_MIGRATION_ID, item_type=t.uint8_t
+            )
+        except KeyError:
+            migration_id = 0
+
+        initial_migration_id = migration_id
+
+        # Migration 1: empty `ADDRMGR` entries are version-dependent and were improperly
+        #              written for CC253x devices.
+        #
+        #              This migration is stateless and can safely be run more than once:
+        #              the only downside is that startup times increase by 10s on newer
+        #              coordinators, which is why the migration ID is persisted.
+        if migration_id < 1:
+            try:
+                entries = await security.read_addr_manager_entries(self._znp)
+            except KeyError:
+                pass
+            else:
+                fixed_entries = []
+
+                for entry in entries:
+                    if entry.extAddr != t.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"):
+                        fixed_entries.append(entry)
+                    elif self._znp.version == 3.30:
+                        fixed_entries.append(const.EMPTY_ADDR_MGR_ENTRY_ZSTACK3)
+                    else:
+                        fixed_entries.append(const.EMPTY_ADDR_MGR_ENTRY_ZSTACK1)
+
+                if entries != fixed_entries:
+                    LOGGER.warning(
+                        "Repairing %d invalid empty address manager entries (total %d)",
+                        sum(i != j for i, j in zip(entries, fixed_entries)),
+                        len(entries),
+                    )
+                    await security.write_addr_manager_entries(self._znp, fixed_entries)
+
+            migration_id = 1
+
+        if initial_migration_id != migration_id:
+            await self._znp.nvram.osal_write(
+                OsalNvIds.ZIGPY_ZNP_MIGRATION_ID, t.uint8_t(migration_id), create=True
+            )
+            await self._znp.reset()
+
     async def _write_stack_settings(self, *, reset_if_changed: bool) -> None:
         """
         Writes network-independent Z-Stack settings to NVRAM.
@@ -1253,7 +1138,10 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                 DeviceId=zigpy.profiles.zha.DeviceType.IAS_CONTROL,
                 DeviceVersion=0b0000,
                 LatencyReq=c.af.LatencyReq.NoLatencyReqs,
-                InputClusters=[clusters.general.Ota.cluster_id],
+                InputClusters=[
+                    clusters.general.Basic.cluster_id,
+                    clusters.general.Ota.cluster_id,
+                ],
                 OutputClusters=[
                     clusters.security.IasZone.cluster_id,
                     clusters.security.IasWd.cluster_id,
@@ -1269,35 +1157,35 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                 DeviceId=zigpy.profiles.zll.DeviceType.CONTROLLER,
                 DeviceVersion=0b0000,
                 LatencyReq=c.af.LatencyReq.NoLatencyReqs,
-                InputClusters=[],
+                InputClusters=[clusters.general.Basic.cluster_id],
                 OutputClusters=[],
             ),
             RspStatus=t.Status.SUCCESS,
         )
 
-    async def _load_network_info(self) -> None:
+    async def load_network_info(self, *, load_devices=False) -> None:
         """
-        Loads low-level network information from NVRAM.
+        Loads network information from NVRAM.
         """
 
-        await self._znp.load_network_info()
+        await self._znp.load_network_info(load_devices=load_devices)
 
-        self.ieee = self._znp.network_info.ieee
-        self.nwk = self._znp.network_info.nwk
-        self.state.network_information.channel = self._znp.network_info.channel
-        self.state.network_information.channel_mask = self._znp.network_info.channels
-        self.state.network_information.pan_id = self._znp.network_info.pan_id
-        self.state.network_information.extended_pan_id = (
-            self._znp.network_info.extended_pan_id
+        self.state.node_information = self._znp.node_info
+        self.state.network_information = self._znp.network_info
+
+    async def write_network_info(
+        self,
+        *,
+        network_info: zigpy.state.NetworkInformation,
+        node_info: zigpy.state.NodeInfo,
+    ) -> None:
+        """
+        Writes network and node state to NVRAM.
+        """
+
+        return await self._znp.write_network_info(
+            network_info=network_info, node_info=node_info
         )
-        self.state.network_information.nwk_update_id = (
-            self._znp.network_info.nwk_update_id
-        )
-        nwk_key = zigpy.state.Key(
-            key=self._znp.network_info.network_key,
-            seq=self._znp.network_info.network_key_seq,
-        )
-        self.state.network_information.network_key = nwk_key
 
     def _find_endpoint(self, dst_ep: int, profile: int, cluster: int) -> int:
         """
@@ -1627,6 +1515,18 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                         if tried_last_good_route and force_relays is not None:
                             force_relays = None
 
+                        # If we fail to contact the device with its IEEE address, don't
+                        # try again.
+                        if (
+                            tried_ieee_address
+                            and dst_addr.mode == t.AddrMode.IEEE
+                            and device is not None
+                        ):
+                            dst_addr = t.AddrModeAddress(
+                                mode=t.AddrMode.NWK,
+                                address=device.nwk,
+                            )
+
                         # Child aging is disabled so if a child switches parents from
                         # the coordinator to another router, we will not be able to
                         # re-discover a route to it. We have to manually drop the child
